"""Language package for Tcl"""

from xpcom import components, ServerException
from koLanguageServiceBase import *
import os

import autoinsert

class koTclLanguage(KoLanguageBase):
    name = "Tcl"
    _reg_desc_ = "%s Language" % name
    _reg_contractid_ = "@activestate.com/koLanguage?language=%s;1" \
                       % (name)
    _reg_clsid_ = "{491A0CE5-7180-425b-A27A-9EA36BCBA50F}"

    namedBlockRE = "^[ \t;]*(proc|class|method)\s+[\w:]+(.*\{\n)?"
    namedBlockDescription = 'Tcl procedures and [incr Tcl] classes'
    defaultExtension = ".tcl"

    # XXX read from config somewhere, download from ActiveState?
    downloadURL = 'http://www.ActiveState.com/Products/ActiveTcl'
    commentDelimiterInfo = { "line": [ "#" ]  }
    variableIndicators = '$'
    supportsSmartIndent = "brace"
    _dedenting_statements = [u'error', u'return', u'break', u'continue']

    sample = """proc loadFile { } {
    # a simple comment
    global f
    global f2
    global fileToOpen
    global fileOpened
    if [file exists $fileToOpen] {
        $f2.text delete 1.0 end
        set ff [open $fileToOpen]
        while {![eof $ff]} {
            $f2.text insert end [read $ff 1000]
        }
        close $ff
        set fileOpened $fileToOpen
    } else {
        $f.entry insert end " does not exist"
    }
}
"""

    styleStdin = components.interfaces.ISciMoz.SCE_TCL_STDIN
    styleStdout = components.interfaces.ISciMoz.SCE_TCL_STDOUT
    styleStderr = components.interfaces.ISciMoz.SCE_TCL_STDERR

    def get_lexer(self):
        if self._lexer is None:
            self._lexer = KoLexerLanguageService()
            self._lexer.setLexer(components.interfaces.ISciMoz.SCLEX_TCL)
            self._lexer.setKeywords(0, self._tcl_keywords)
            self._lexer.supportsFolding = 1
        return self._lexer


    def get_linter(self):
        if self._linter is None:
            self._linter = KoLinterLanguageService("@activestate.com/koTclLinter;1")
        return self._linter

    def get_completer(self):
        if self._completer is None:
            self._completer = components.classes["@activestate.com/koTclCompletionLanguageService;1"].getService(components.interfaces.koICompletionLanguageService)
        return self._completer

    def get_interpreter(self):
        if self._interpreter is None:
            self._interpreter = components.classes["@activestate.com/koTclInfoEx;1"].getService()
        return self._interpreter
    


    # XXX later these need to be centralized.

    _tcl_keywords = ["after", "append", "array", "auto_execok",
                     "auto_import", "auto_load", "auto_load_index",
                     "auto_qualify", "bgerror", "binary", "break", "case",
                     "catch", "cd", "clear", "clock", "close", "concat",
                     "continue", "else", "then", "elseif", "encoding",
                     "eof", "error", "eval", "exec", "exit", "expr",
                     "fblocked", "fconfigure", "fcopy", "file",
                     "fileevent", "flush", "for", "foreach", "format",
                     "gets", "glob", "global", "history", "if", "incr",
                     "info", "interp", "join", "lappend", "lindex",
                     "linsert", "list", "llength", "load", "lrange",
                     "lreplace", "lsearch", "lsort", "namespace", "open",
                     "package", "pid", "pkg_compareExtension",
                     "pkg_mkIndex", "proc", "puts", "pwd", "read",
                     "regexp", "regsub", "rename", "return", "scan",
                     "seek", "set", "socket", "source", "split", "string",
                     "subst", "switch", "tcl_unknown", "tell", "time",
                     "trace", "unknown", "unset", "update", "uplevel",
                     "upvar", "variable", "vwait", "while",
                     "bell", "bind", "bindtags", "button", "canvas",
                     "checkbutton", "clipboard", "destroy", "entry",
                     "event", "focus", "font", "frame", "grab", "grid",
                     "image", "label", "listbox", "lower", "menu",
                     "menubutton", "message", "option", "pack", "place",
                     "radiobutton", "raise", "registry", "scale",
                     "scrollbar", "selection", "spinbox",
                     "tcl_findLibrary", "text", "tk", "tk_chooseColor",
                     "tk_chooseDirectory", "tk_getOpenFile",
                     "tk_getSaveFile", "tk_menuSetFocus", "tk_messageBox",
                     "tk_popup", "tk_textCopy", "tk_textCut",
                     "tk_textPaste", "tkwait", "toplevel", "winfo", "wm"]
