# Copyright (c) 2000-2003 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

# A better implementation of uri's than urllib stuff

# needed for ftpfile
import urlparse, urllib
import stat, os, sys, copy, md5
import types

from xpcom import components, COMException, ServerException, nsError

def addSchemeToParser(scheme):
    urlparse.uses_relative.append(scheme)
    urlparse.uses_netloc.append(scheme)
    urlparse.uses_params.append(scheme)
    urlparse.uses_query.append(scheme)
    urlparse.uses_fragment.append(scheme)

# Need to do this so we can correctly parse
# uri's given to us by the debugger
addSchemeToParser('kodebugger')
addSchemeToParser('koremote')
addSchemeToParser('dbgp')

class URILibError(Exception):
    pass


class URIParser(object):
    def __init__(self, uri=None):
        self._clear()
        if uri:
            self._setFileNames(uri)
    
    def _clear(self):
        self.fileName = None
        self._uri = None
        self._fileParsed = ['','','','','']
        self._baseName = None
        self._dirName = None
        self._ext = None
    
    def dump(self):
        print
        print "_uri:     [%s]"%self._uri
        print "fileName: [%s]"%self.fileName
        print "URI:      [%s]"%self.URI
        print "scheme:   [%s]"%self.scheme
        print "path:     [%s]"%self.path
        print "leafName: [%s]"%self.leafName
        print "baseName: [%s]"%self.baseName
        print "dirName:  [%s]"%self.dirName
        print "ext:      [%s]"%self.ext
        print
        
    def _parseURI(self, uri):
        uri = uri.replace('\\','/')
        
        # fix the uri if we get the lame pipe in place of colon uri's
        uri = uri.replace('|',':')
        
        # if we got a netscape unc file uri, fix it
        if uri.find('file://///')==0:
            return self._parseURI(uri[8:])
            
        # change c:/ to /c:/ so urlsplit works right
        if uri[1] == ':':
            uri = '/'+uri
            
        parts = list(urlparse.urlsplit(uri,'file',0))
        
        # prevent /c:/ paths from persisting
        if parts[2].find(':') >= 0:
            parts[2] = parts[2][1:]

        # fix the UNC file uri  this assums netloc in a file uri
        # is a machine name, thus this is a UNC
        if parts[0]=='file' and parts[1]:
            parts[2]='//'+parts[1]+parts[2]
            parts[1]=''
            
        return parts
    
    def _buildURI(self, parts):
        # XXX we may need to do more about quoting
        # if a windows path, fix it
        uparts = copy.copy(parts)
        uparts[2] = uparts[2].replace('\\','/')
        if uparts[2].find('%') >= 0:
             uparts[2] = urllib.unquote(uparts[2])
        if uparts[2].find(' ') >= 0:
            if uparts[2].find(':') == 1:
                uparts[2] = uparts[2][:2] + urllib.quote(uparts[2][2:])
            else:
                # XXX a bit of a hack to support koremote
                # which contains a sub-url.  currently the only
                # koremote is our ftp support, which doesn't expect
                # to get quoted url's
                moreparts = self._parseURI(uparts[2])
                if not moreparts:
                    uparts[2] = urllib.quote(uparts[2])
        uparts[2] = uparts[2].replace('\\','/')
        return urlparse.urlunsplit(tuple(uparts))

    def _setFileNames(self, fileName):
        self._clear()
        if fileName:
            if fileName.find('%') != -1:
                fileName = urllib.unquote(fileName)
            self.fileName = fileName
            self._fileParsed = self._parseURI(fileName)
            self._uri = self._buildURI(self._fileParsed)

    #attribute string URI; 
    def get_URI(self):
        return self._uri
    
    def set_URI(self, uri):
        if uri is None:
            self._clear()
        else:
            self._setFileNames(uri)
    URI = property(get_URI,set_URI)

    def get_displayPath(self):
        if self.scheme == 'file':
            return self.get_path() 
        else:
            return self.get_URI()
    
    #attribute string path; 
    def get_path(self):
        path = os.path.normpath(self._fileParsed[2])
        path = urllib.unquote(path)
        return path 

    def set_path(self, path):
        if path.find('://') >= 0:
            self.set_URI(path)
        else:
            self._fileParsed[2] = os.path.normpath(path)
            self._setFileNames(self._buildURI(self._fileParsed))
    path = property(get_path,set_path)

    #attribute string leafName; 
    def get_leafName(self):
        return self.baseName

    def set_leafName(self, leafName):
        self.path = os.path.join(self.dirName,leafName)
    leafName = property(get_leafName,set_leafName)

    #readonly attribute string baseName; 
    def get_baseName(self):
        if not self._baseName:
            if self.path:
                self._baseName = os.path.basename(self.path)
        return self._baseName
    baseName = property(get_baseName)

    #readonly attribute string dirName; 
    def get_dirName(self):
        if not self._dirName:
            if self.path:
                self._dirName = os.path.dirname(self.path)
        return self._dirName
    dirName = property(get_dirName)

    #readonly attribute string scheme; 
    def get_scheme(self):
        return self._fileParsed[0]
    scheme = property(get_scheme)

    #readonly attribute string ext;
    def get_ext(self):
        if not self._ext:
            if self.path:
                self._ext = os.path.splitext(self.path)[1]
        return self._ext
    ext = property(get_ext)

    #readonly attribute string md5name; 
    def get_md5name(self):
        return md5.new(self._uri).hexdigest()
    md5name = property(get_md5name)

class FileHandlerBase(object):
    def __init__(self):
        self.clearstat()
        self.clearscc()
        self._file = None
        self._path = None
        self._mode = 'rb'
        
        # used by file status service
        self.update = 0
        self.dofileupdate = 0
        
        # lastAccessedTime is special, in that it can change
        # but we don't care if it does.  It's only used in ko
        # for display purposes.  If we keep this in the stat
        # cache for this file, it makes checking disk changes
        # more difficult, so instead, we store it seperately
        # see get_hasChanged below
        self.lastAccessedTime = 0
        

    def get_stats(self):
        return self._stats
    stats = property(get_stats)

    def get_scc(self):
        return self._scc

    # this is expected to be set by the filestatusservice
    # which unwraps this object and calls setscc, so
    # this function is not in the idl
    def set_scc(self, scc_info):
        self.clearscc()
        self._scc.update(scc_info)

    scc = property(get_scc, set_scc)

    def get_sccSummary(self):
        retval = '-'.join([self._scc['sccType'],
                         self._scc['sccDirType'],
                         self._scc['sccStatus'],
                         str(self._scc['sccNeedSync']),
                         self._scc['sccAction']])
        return retval

    sccSummary = property(get_sccSummary) # no setter

    def clearstat(self):
        self._stats = {}

    def clearscc(self):
        self._scc = {'sccType':'',
                     'sccDirType': '',   
                     'sccExclude':0, 
                     'sccHaveOnDisk':0, 
                     'sccLocalRevision':'',
                     'sccRevDate':'',
                     'sccDepotRevision':'',
                     'sccNeedSync':0, 
                     'sccSync':0,     
                     'sccOk':0,
                     'sccConflict':0, 
                     'sccAction':'',
                     'sccStatus':'',
                     'sccChange':''}

    def __getattr__(self,name):
        if name in self.stats:
            return self.stats[name]
        if name in self.scc:
            return self.scc[name]
        if self._file is None or not hasattr(self._file,name):
            raise AttributeError, name
        return getattr(self._file,name)

    def getStatusMap(self):
        status = self.stats.copy()
        status.update(self.scc)
        keys = map(str, status.keys())
        values = map(str, status.values())
        return keys,values


class FileHandler(FileHandlerBase):
    isLocal = 1
    
    def __init__(self,path):
        FileHandlerBase.__init__(self)
        uri = URIParser(path)
        if uri.scheme != 'file':
            raise URILibError("Invalid File Scheme: %s" % uri.scheme)
        self._path = uri.path
        
        # get a decoded version of uri.path.  This exists to fix reading
        # files with latin-1 or other 8 bit encodings on systems where
        # utf-8 support is incomplete (bug 29040)
        if type(self._path) == types.UnicodeType:
            initService = components.classes['@activestate.com/koInitService;1'].\
                            getService(components.interfaces.koIInitService);
            encodingServices = components.classes['@activestate.com/koEncodingServices;1'].\
                            getService(components.interfaces.koIEncodingServices);
            self._decodedPath = encodingServices.encode(self._path,
                                                    initService.getStartupEncoding(),
                                                    'strict')
        else:
            self._decodedPath = self._path

    def __del__(self):
        if self._file:
            self.close()

    def close(self):
        if not self._file:
            raise URILibError("file not opened: '%s'" % self._path)
        try:
            try:
                self._file.close()
            except EnvironmentError, ex:
                lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                               .getService(components.interfaces.koILastErrorService)
                lastErrorSvc.setLastError(ex.errno, ex.strerror)
                raise ServerException(nsError.NS_ERROR_FAILURE, ex.strerror)
                return 0
        finally:
            self._file = None
        
    def open(self, mode):
        self._mode = mode
        if self._file is not None:
            raise URILibError("Cannot open a file again without closing it first!")
        if self._stats and self._stats['exists']==0:
            self._stats = None
        try:
            self._file = open(self._decodedPath, self._mode)
        except EnvironmentError, ex:
            lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                           .getService(components.interfaces.koILastErrorService)
            lastErrorSvc.setLastError(ex.errno, ex.strerror)
            raise ServerException(nsError.NS_ERROR_FAILURE, ex.strerror)
            return 0
        return self._file is not None

    def read(self, nBytes):
        try:
            return self._file.read(nBytes)
        except EnvironmentError, ex:
            lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                           .getService(components.interfaces.koILastErrorService)
            lastErrorSvc.setLastError(ex.errno, ex.strerror)
            raise ServerException(nsError.NS_ERROR_FAILURE, ex.strerror)

    def write(self, text):
        try:
            self._file.write(text)
        except EnvironmentError, ex:
            lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                           .getService(components.interfaces.koILastErrorService)
            lastErrorSvc.setLastError(ex.errno, ex.strerror)
            raise ServerException(nsError.NS_ERROR_FAILURE, ex.strerror)

    def __get_stats(self):
        _stats = {'mode':'','ino':'','dev':'','nlink':'',
                    'uid':'','gid':'','fileSize':0,
                    'lastModifiedTime':0,'createdTime':0,'isReadable':0,
                    'isWriteable':0,'isExecutable':0,'isReadOnly':0,
                    'isReadWrite':0,'exists':0,'isDirectory':0,
                    'isFile':0,'isSymlink':0,'isSpecial':0,'permissions':0,
                    'isHidden':0}
        try:
            stats = os.stat(self._decodedPath)
            mode = stats[stat.ST_MODE]
            _stats['mode'] = mode
            _stats['ino'] = stats[stat.ST_INO]
            _stats['dev'] = stats[stat.ST_DEV]
            _stats['nlink'] = stats[stat.ST_NLINK]
            _stats['uid'] = stats[stat.ST_UID]
            _stats['gid'] = stats[stat.ST_GID]
            _stats['fileSize'] = stats[stat.ST_SIZE]
            # see comment on lastAccessedTime above
            self.lastAccessedTime = stats[stat.ST_ATIME]
            _stats['lastModifiedTime'] = stats[stat.ST_MTIME]
            _stats['createdTime'] = stats[stat.ST_CTIME]

            _stats['isReadable'] = mode & stat.S_IREAD == stat.S_IREAD
            _stats['isWriteable'] = mode & stat.S_IWRITE == stat.S_IWRITE
            _stats['isExecutable'] = mode & stat.S_IEXEC == stat.S_IEXEC
            _stats['isDirectory'] = stat.S_ISDIR(mode)
            _stats['isFile'] = stat.S_ISREG(mode)
            _stats['isSymlink'] = stat.S_ISLNK(mode)
            _stats['isSpecial'] = stat.S_ISCHR(mode) or stat.S_ISBLK(mode) or stat.S_ISFIFO(mode) or stat.S_ISSOCK(mode)
            _stats['permissions'] = stat.S_IMODE(mode)
            
            _stats['isReadOnly'] = _stats['isReadable'] and not _stats['isWriteable']
            _stats['isReadWrite'] = _stats['isReadable'] and _stats['isWriteable']
            _stats['exists']=1
            #print repr(_stats)
        except OSError,e:
            if e.errno != 2:
                raise
        return _stats

    def get_stats(self):
        if not self._stats:
            self._stats = self.__get_stats()
        return self._stats
    stats = property(get_stats)

    def get_hasChanged(self):
        if not self._stats:
            return 0
        tmpstats = self.__get_stats()
        if self._stats != tmpstats:
            self._stats = tmpstats
            self.update = 0
            self.dofileupdate = 1
            return 1
        return 0
    hasChanged = property(get_hasChanged)


class URIHandler(FileHandlerBase):
    isLocal = 0
    
    def __init__(self,path):
        import urllib2
        FileHandlerBase.__init__(self)
        uri = URIParser(path)
        if uri.scheme == 'file':
            raise URILibError("Invalid File Scheme: %s" % uri.scheme)
        self._path = uri.URI
    
    def __del__(self):
        if self._file:
            self.close()

    def close(self):
        if not self._file:
            raise URILibError("file not opened: '%s'" % self._path)
        self._file.close()
        self._file = None
        
    def open(self, mode):
        import urllib2
        self._mode = mode
        if self._file is not None:
            raise URILibError("Cannot open a file again without closing it first!")
        if self._stats and self._stats['exists']==0:
            self._stats = None
        self._file = urllib2.urlopen(self._path)
        return self._file is not None

    def __get_stats(self):
        _stats = {'mode':'','ino':'','dev':'','nlink':'',
                    'uid':'','gid':'','fileSize':0,
                    'lastModifiedTime':0,'createdTime':0,'isReadable':1,
                    'isWriteable':0,'isExecutable':0,'isReadOnly':1,
                    'isReadWrite':0,'exists':1,'isDirectory':0,
                    'isFile':0,'isSymlink':0,'isSpecial':0,'permissions':0,
                    'isHidden':0}
        # XXX need to implement this stuff for ftp/http
        return _stats
    
    def get_stats(self):
        if not self._stats:
            self._stats = self.__get_stats()
        return self._stats
    stats = property(get_stats)

    def get_hasChanged(self):
        return 0
    hasChanged = property(get_hasChanged)


class StartPageHandler(FileHandlerBase):
    #XXX Shane, should this inherit from FileHandlerBase, because no chrome
    #    file should use any of SCC status stuff?

    isLocal = 0
    
    def __init__(self, path):
        FileHandlerBase.__init__(self)
        if path != "chrome://jaguar/content/startpage/startpage.xml#view-startpage":
            raise URILibError("Invalid Start Page path: %s" % path)
    
    def __del__(self):
        pass

    def close(self):
        pass
        
    def open(self, mode):
        return 0

    def __get_stats(self):
        _stats = {'mode':'','ino':'','dev':'','nlink':'',
                  'uid':'','gid':'','fileSize':0,
                  'lastModifiedTime':0,'createdTime':0,'isReadable':1,
                  'isWriteable':0,'isExecutable':0,'isReadOnly':1,
                  'isReadWrite':0,'exists':1,'isDirectory':0,
                  'isFile':1,'isSymlink':0,'isSpecial':0,'permissions':0,
                  'isHidden':0}
        # XXX need to implement this stuff from nsIFile stuff
        return _stats
    
    def get_stats(self):
        if not self._stats:
            self._stats = self.__get_stats()
        return self._stats
    stats = property(get_stats)

    def get_hasChanged(self):
        return 0
    hasChanged = property(get_hasChanged)

    def get_displayPath(self):
        """A HACK to override the start page view's document.file.displayPath
        until http://bugs.activestate.com/Komodo/show_bug.cgi?id=25526 is
        implemented.
        """
        return "Start Page"

    
class xpURIHandler(FileHandlerBase):
    isLocal = 0
    
    def __init__(self,path):
        # XXX need to use URIFile
        import urllib2
        FileHandlerBase.__init__(self)
        uri = URIParser(path)
        self._path = path
    
    def __del__(self):
        if self._file:
            self.close()

    def close(self):
        if not self._file:
            raise URILibError("file not opened: '%s'" % self._path)
        self._file.close()
        self._file = None
        
    def open(self, mode):
        from xpcom.file import URIFile
        self._mode = mode
        if self._file is not None:
            raise URILibError("Cannot open a file again without closing it first!")
        if self._stats and self._stats['exists']==0:
            self._stats = None
        try:
            self._file = URIFile(self._path)
        except Exception, e:
            # Mozilla could not open the file, so it either does not exist
            # or some other error which we unfortunately do not get good info
            # on from mozilla.  One example is a dbgp uri being restored
            # on startup.  We pass here and let the file be non-existent
            self._file = None
            if self._stats:
                self._stats['exists'] = 0
                self._stats['isReadable'] = 0
                self._stats['isReadOnly'] = 0
        return self._file is not None

    def __get_stats(self):
        # exists must be true for an open file attempt to be made
        # we set isSpecial to true here because this class is only used for
        # mozilla based uri's, such as the dbgp:// implementation or chrome
        # uri's.  We do not want these uri's restored when Komodo restarts.
        # this way we can genericaly avoid trying to restore these.
        _stats = {'mode':'','ino':'','dev':'','nlink':'',
                    'uid':'','gid':'','fileSize':'',
                    'lastModifiedTime':0,'createdTime':0,'isReadable': 1,
                    'isWriteable':0,'isExecutable':0,'isReadOnly': 1,
                    'isReadWrite':0,'exists': 1,'isDirectory':0,
                    'isFile':0,'isSymlink':0,'isSpecial':1,'permissions':0,
                    'isHidden':0}
        # XXX need to implement this stuff for ftp/http
        return _stats
    
    def get_stats(self):
        if not self._stats:
            self._stats = self.__get_stats()
        return self._stats
    stats = property(get_stats)

    def get_hasChanged(self):
        return 0
    hasChanged = property(get_hasChanged)

class RemoteURIHandler(FileHandlerBase):
    isLocal = 0
    
    def __init__(self,path):
        # XXX need to use URIFile
        import urllib2, re
        FileHandlerBase.__init__(self)
        uri = URIParser(path)
        if uri.scheme != 'koremote':
            raise URILibError("Invalid File Scheme: %s" % uri.scheme)
        self._url_parts = re.match(r'^(.*?)://(.*?)/(.*)$',
                             path).groups()
        self._server_alias = self._url_parts[1]
        self._path = self._url_parts[2]
        self._mode = 'r'

    def __del__(self):
        if self._file:
            self.close()

    def close(self):
        if not self._file:
            raise URILibError("file not opened: '%s'" % self._path)
        self._file.close()
    
    def _initFTP(self):
        if not self._file:
            self._file = components.classes["@activestate.com/koFTPFile;1"]\
                     .createInstance(components.interfaces.koIFTPFile)
            self._file.initServer(self._path, self._mode, self._server_alias)
        
    def open(self, mode):
        if mode != self._mode:
            self._file = None
        self._mode = mode
        self._stats = None
        self._initFTP()
        return self._file is not None

    def __get_stats(self):
        # XXX need to implement this stuff for ftp/http
        _stats = {'mode':'','ino':'','dev':'','nlink':'',
                    'uid':'','gid':'','fileSize':0,
                    'lastModifiedTime':0,'createdTime':0,'isReadable':1,
                    'isWriteable':0,'isExecutable':0,'isReadOnly':1,
                    'isReadWrite':0,'exists':1,'isDirectory':0,
                    'isFile':0,'isSymlink':0,'isSpecial':0,'permissions':0,
                    'isHidden':0}
        if self._file:
            try:
                self._file.ftpfile.refresh()
                _stats['fileSize']          = self._file.ftpfile.fileSize
                _stats['lastModifiedTime']  = self._file.ftpfile.lastModifiedTime
                _stats['createdTime']       = self._file.ftpfile.lastModifiedTime
                self.lastAccessedTime       = self._file.ftpfile.lastModifiedTime
                _stats['isReadable']        = self._file.ftpfile.isReadable()
                _stats['isWriteable']       = self._file.ftpfile.isWritable()
                _stats['isExecutable']      = self._file.ftpfile.isExecutable()
                _stats['isReadOnly']        = _stats['isReadable'] and not _stats['isWriteable']
                _stats['isReadWrite']       = _stats['isReadable'] and _stats['isWriteable']
                _stats['exists']            = self._file.ftpfile.exists()
                _stats['isDirectory']       = self._file.ftpfile.isDirectory()
                _stats['isFile']            = self._file.ftpfile.isFile()
                _stats['isSymlink']         = self._file.ftpfile.isSymlink()
                _stats['isSpecial']         = self._file.ftpfile.isSpecial()
                _stats['permissions']       = self._file.ftpfile.permissions
                _stats['isHidden']          = self._file.ftpfile.isHidden()
            except Exception, e:
                print e
        return _stats
    
    def get_stats(self):
        if not self._stats:
            self._stats = self.__get_stats()
        return self._stats
    stats = property(get_stats)

    def get_hasChanged(self):
        if not self._stats:
            return 0
        tmpstats = self.__get_stats()
        if self._stats != tmpstats:
            self._stats = tmpstats
            self.update = 0
            self.dofileupdate = 1
            return 1
        return 0
    hasChanged = property(get_hasChanged)
